
/*
Copyright 2011 Aurélien Bellet
aurelien.bellet@univ-st-etienne.fr

This file is part of GESL.

GESL is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GESL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GESL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GESL_classify.h"


// build Levenshtein similarity matrix(ces) from training data
// there is one matrix per binary problem
double*** build_lev_mat (sample *data) {

  int i,j,k,l,ind,sizei,sizej;
  double ***sim_mat;
  int nb_class = data->nb_class;

  sim_mat = (double ***) malloc(nb_class*nb_class*sizeof(double**));

  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      ind = i*nb_class+j;
      sim_mat[ind] = (double**) malloc((data->class[i]->size+data->class[j]->size)*sizeof(double*));
      for (k=0; k<data->class[i]->size+data->class[j]->size; k++) {
	sim_mat[ind][k] = (double *) malloc((data->class[i]->size+data->class[j]->size)*sizeof(double));
      }
    }
  }

  fprintf(stderr, "Computing Levenshtein similarity matrix(ces) from training data... ");

  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      fprintf(stderr, "(%d,%d) ",i,j);
      ind = i*nb_class+j;
      sizei = data->class[i]->size;
      sizej = data->class[j]->size;
      for (k=0; k<sizei+sizej; k++) {
	for (l=k; l<sizei+sizej; l++) {
	  if (k < sizei) {
	    if (l < sizei)
	      sim_mat[ind][k][l] = -levenshtein(data->class[i]->t[k],data->class[i]->t[l]) * 1.0;
	    else
	      sim_mat[ind][k][l] = -levenshtein(data->class[i]->t[k],data->class[j]->t[l-sizei]) * 1.0;
	  } else {
	    if (l < data->class[i]->size)
	      sim_mat[ind][k][l] = -levenshtein(data->class[j]->t[k-sizei],data->class[i]->t[l]) * 1.0;
	    else
	      sim_mat[ind][k][l] = -levenshtein(data->class[j]->t[k-sizei],data->class[j]->t[l-sizei]) * 1.0;
	  }
	  sim_mat[ind][l][k] = sim_mat[ind][k][l];
	}
      }
    }
  }

  fprintf(stderr,"Done\n");
  return sim_mat;
 
}


// return the (mean,std) values for each matrix
// used later to normalize the Levenshtein distance to [-1,1]
double** normalize_mat (double*** sim_mat, sample *data) {
  int i,j,k,l,ind,sizei,sizej;
  double mean,std,**norm;
  int nb_class = data->nb_class;

  norm = (double **) malloc(nb_class*nb_class*sizeof(double *));
  for (i=0; i<nb_class*nb_class; i++) {
    norm[i] = (double *) malloc(2*sizeof(double));
  }

  fprintf(stderr, "Normalizing similarity matrix(ces)... ");
  
  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      ind = i*nb_class+j;
      sizei = data->class[i]->size;
      sizej = data->class[j]->size;
      mean = 0;
      std = 0;
      for (k=0; k<sizei+sizej; k++)
	for (l=0; l<sizei+sizej; l++)
	  mean += sim_mat[ind][k][l];
      mean /= (sizei+sizej)*(sizei+sizej);
      for (k=0; k<sizei+sizej; k++)
	for (l=0; l<sizei+sizej; l++)
	  std += pow(sim_mat[ind][k][l] - mean,2);
      std /= (sizei+sizej)*(sizei+sizej);
      std = sqrt(std);
      norm[ind][0] = mean;
      norm[ind][1] = std;
    }
  }
  fprintf(stderr,"Done\n");
  return norm;
}


// load edit matrix(ces) C learned with GESL
double*** load_matrices (char *matrix_dir, int alph_size, int nb_class) {
  FILE *f;
  char line[1000];
  char path[1000];
  int i,j,p,q,ind;
  double ***c;

  c = (double ***) malloc(nb_class*nb_class*sizeof(double **));
  for (i=0; i<nb_class*nb_class; i++) {
    c[i] = (double **) malloc((alph_size+1)*sizeof(double *));
    for (j=0; j<alph_size+1; j++)
      c[i][j] = (double *) malloc((alph_size+1)*sizeof(double));
  }

  for (p=0; p<nb_class-1; p++) {
    for (q=p+1; q<nb_class; q++) {
      ind = p*nb_class+q;
      sprintf(path,"%s/matrix_%d-%d",matrix_dir,p,q);
      f = fopen(path,"rt");
      if (f == NULL) {
	printf("File %s does not exist.\n",path);
	exit(-1);
      }
      for (i=0; i<alph_size+1; i++)
        for (j=0; j<alph_size+1; j++)
          fscanf(f,"%lf ",&c[ind][i][j]);
      fclose(f);
    }
  }

  return c;

}


// compute the similarity with respect to matrix c learned with GESL
// it first computes the Levenshtein script using script_lev() from "GESL.h"
// and applies custom costs c to it
double K_G (char *x, char *y, int alph_size, char ref_char, double **c) {

  int i,j;
  int **script = script_lev(x,y,alph_size,ref_char);
  double d;
  int script_size = 0;

  d = 0;
  for (i=0; i<=alph_size; i++) {
    for (j=0; j<=alph_size; j++) {
      d += script[i][j]*c[i][j];
      script_size += script[i][j];
    }
  }

  for (i=0; i<=alph_size; i++)
    free(script[i]);
  free(script);
  
  if (script_size == 0)
    return 1;

  return 2*exp(-d)-1;
  
}


// build GESL similarity matrix(ces) from training data
double*** build_GESL_mat (sample *data, int alph_size, char ref_char, double ***C) {

  int i,j,k,l,ind,sizei,sizej;
  double ***sim_mat;
  int nb_class = data->nb_class;

  sim_mat = (double ***) malloc(nb_class*nb_class*sizeof(double**));

  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      ind = i*nb_class+j;
      sim_mat[ind] = (double**) malloc((data->class[i]->size+data->class[j]->size)*sizeof(double*));
      for (k=0; k<data->class[i]->size+data->class[j]->size; k++) {
	sim_mat[ind][k] = (double *) malloc((data->class[i]->size+data->class[j]->size)*sizeof(double));
      }
    }
  }

  fprintf(stderr, "Computing GESL similarity matrix(ces) from training data... ");

  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      fprintf(stderr, "(%d,%d) ",i,j);
      ind = i*nb_class+j;
      sizei = data->class[i]->size;
      sizej = data->class[j]->size;
      for (k=0; k<sizei+sizej; k++) {
	for (l=0; l<sizei+sizej; l++) {
	  if (k < sizei) {
	    if (l < sizei)
	      sim_mat[ind][k][l] = K_G(data->class[i]->t[k],data->class[i]->t[l],alph_size,ref_char,C[ind]);
	    else
	      sim_mat[ind][k][l] = K_G(data->class[i]->t[k],data->class[j]->t[l-sizei],alph_size,ref_char,C[ind]);
	  } else {
	    if (l < sizei)
	      sim_mat[ind][k][l] = K_G(data->class[j]->t[k-sizei],data->class[i]->t[l],alph_size,ref_char,C[ind]);
	    else
	      sim_mat[ind][k][l] = K_G(data->class[j]->t[k-sizei],data->class[j]->t[l-sizei],alph_size,ref_char,C[ind]);
	  }
	}
      }
    }
  }

  fprintf(stderr,"Done\n");
  return sim_mat;
 
}


// build linear program(s) in MOSEK LP format
void build_lp_mosek (sample *train, double ***sim_mat, double **norm, char* data_dir, int sim, double lambda) {
  
  int i,j,k,m,ind,sizek,sizem;
  double temp,mean,std;
  FILE *f;
  char path[1000];

  fprintf(stderr, "Building LP(s)... ");

  for (k=0; k<(train->nb_class)-1; k++) {
    for (m=k+1; m<train->nb_class; m++) {
      fprintf(stderr,"(%d,%d) ",k,m);
      ind = k*train->nb_class+m;
      
      // if using Levenshtein, need to normalize to N(0,1) using std and mean
      if (sim == 0) {
        mean = norm[k*train->nb_class+m][0];
        std = norm[k*train->nb_class+m][1];
      // if using similarity learned with GESL, no need for normalization
      } else if (sim == 1) {
        mean = 0;
        std = 1;
      }

      sizek = train->class[k]->size;
      sizem = train->class[m]->size;

      sprintf(path,"%s/model_%d-%d-%d-%E.lp",data_dir,k,m,sim,lambda);
      f = fopen(path,"wt");
      if (f == NULL) {
	fprintf(stderr,"File %s does not exist.\n",path);
	exit(-1);
      }

      // objective function
      fprintf(f,"minimize\n");
      for (i=0; i<sizek+sizem; i++)
	fprintf(f,"xi_%d + ",i);
      for (j=0; j<sizek+sizem-1; j++)
	fprintf(f,"%E alpha_plus_%d + %E alpha_minus_%d + ",lambda,j,lambda,j);
      fprintf(f,"%E alpha_plus_%d + %E alpha_minus_%d\n",lambda,sizek+sizem-1,lambda,sizek+sizem-1);

      fprintf(f,"subject to\n");

      // constraints \xi_i >= 1 - \sum \alpha_j*l(x_i)*K(x_i,x'_j)
      for (i=0; i<sizek+sizem; i++) {
	fprintf(f,"xi_%d",i);
	for (j=0; j<sizek+sizem; j++) {
	  if (i < sizek)
	    temp = thresh((sim_mat[ind][i][j]-mean)/std);
          else
            temp = -thresh((sim_mat[ind][i][j]-mean)/std);
	  if (temp > 0)
	    fprintf(f," + %E alpha_plus_%d - %E alpha_minus_%d",temp,j,temp,j);
	  else if (temp < 0)
	    fprintf(f," - %E alpha_plus_%d + %E alpha_minus_%d",-temp,j,-temp,j);
	}
	fprintf(f," >= 1\n");
      }

      fprintf(f,"end\n");

      fclose(f);
    }
  }
  fprintf(stderr, "Done\n");
}


// learn the classifier(s) alpha by solving the LP problems with MOSEK
// saves the classifier(s) to data_dir
double** solve_lp_mosek (char* data_dir, int nb_class, sample *train, int sim, double lambda) {
  int i,j,k,ind_i,d;
  FILE *f;
  char path[1000],errname[1000],errdesc[1000];
  MSKtask_t task;
  MSKenv_t env;
  MSKrescodee res;
  double **alpha,*var;
  int n = nb_class*nb_class;
  alpha = (double **) malloc(n*sizeof(double*));

  fprintf(stderr, "Solving LP(s)... ");

  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      fprintf(stderr,"(%d,%d) ",i,j);
      ind_i = i*nb_class+j;

      d = train->class[i]->size+train->class[j]->size;

      alpha[ind_i] = (double*) malloc(d*sizeof(double));

      task = NULL;
      env = NULL;
      res = MSK_RES_OK; 
      res = MSK_makeenv(&env, NULL, NULL, NULL, NULL);
      if (res==MSK_RES_OK)
	res = MSK_initenv (env);
      if (res==MSK_RES_OK)
	res = MSK_maketask(env,d*3,d,&task); 
      sprintf(path,"%s/model_%d-%d-%d-%E.lp",data_dir,i,j,sim,lambda);
      if (res==MSK_RES_OK)
	res = MSK_readdata(task,path); 
      if (res==MSK_RES_OK)
	res = MSK_optimize(task);

      if (res != MSK_RES_OK) {
        // in case of a MOSEK error, print error code and description
        printf("\nMOSEK: An error occurred while optimizing.\n");
        MSK_getcodedesc(res,errname,errdesc);
        printf("Error %s - '%s'\n",errname,errdesc);
        exit(-1);
      } 

      /*
      // print the final objective function value
      double primalobj;
      MSK_getprimalobj (task,MSK_SOL_BAS,&primalobj);
      fprintf(stderr,"Objective function: %f\n",primalobj);
      */

      var = (double *) malloc(d*2*sizeof(double));;
      res = MSK_getsolutionslice(task,MSK_SOL_BAS,MSK_SOL_ITEM_XX,d,d*3,var);

      MSK_deletetask(&task);
      MSK_deleteenv(&env);

      sprintf(path,"%s/model_%d-%d-%d-%E",data_dir,i,j,sim,lambda);
      f = fopen(path,"wt");
      if (f == NULL) {
	fprintf(stderr,"File %s does not exist.\n",path);
	exit(-1);
      }

      for (k=0; k<d*2; k++,k++) {
	alpha[ind_i][k/2] = var[k]-var[k+1];
	if (alpha[ind_i][k/2] != 0)
	  fprintf(f,"%d %lf\n",k/2,alpha[ind_i][k/2]);
      }

      fclose(f);

      free(var);

      sprintf(path,"rm -f %s/model_%d-%d-%d-%E.lp",data_dir,i,j,sim,lambda);
      system(path);

    }
  }
  fprintf(stderr, "Done\n");
  return alpha;
}


// print the size of the learned classifier(s) alpha
void classifier_size (sample *train, double **alpha) {
  int i,j,k,m,ind,size_til;
  int total_r,total_unique,size_total;
  int r,*alpha2;
  int nb_class = train->nb_class;

  //fprintf(stderr,"Size of classifiers (nb of reasonable points): ");

  size_total = 0;
  for (i=0; i<nb_class; i++)
    size_total += train->class[i]->size;

  alpha2 = (int *) calloc(size_total,sizeof(int));

  total_r = 0; total_unique = 0; 
  for (i=0; i<nb_class-1; i++) {
    for (j=i+1; j<nb_class; j++) {
      r = 0;
      ind = i*nb_class+j;
      for (k=0; k<train->class[i]->size+train->class[j]->size; k++) {
	if (alpha[ind][k] != 0) {
	  r++;
	  if (k < train->class[i]->size) {
	    size_til = 0;
	    for (m=0; m<i; m++)
	      size_til += train->class[m]->size;
	    alpha2[size_til+k] = 1;
	  } else {
	    size_til = 0;
	    for (m=0; m<j; m++)
	      size_til += train->class[m]->size;
	    alpha2[size_til+k-train->class[i]->size] = 1;
          }
	}
      }
      //fprintf(stderr,"(%d,%d): %d; ",i,j,r);
      total_r += r;
    }
  }

  for (i=0; i<size_total; i++) {
    if (alpha2[i] == 1)
      total_unique++;
  }

  fprintf(stderr,"Average size of binary classifiers: %.2f\n",total_r/(nb_class*(nb_class-1)/2.0));
  fprintf(stderr,"Nb of unique reasonable points: %d\n",total_unique);

  free(alpha2);
}


// classify test data according to learned classifier(s) alpha
// print confusion matrix and accuracy
void classify_test (sample *train, sample *test, int nb_class, double **norm, double **alpha, double ***C, int alph_size, char ref_char, int sim) {

  fprintf(stderr,"Classifying test data...");

  int i,j,k,l,m,ind,sizelk,sizell,pred,nb_errors,sizet_total;
  int **confusion_matrix;
  double mean,std,value,tmp,max,*vote;

  nb_errors = 0;
  sizet_total = 0;

  confusion_matrix = (int **) calloc(nb_class,sizeof(int*));
  for (i=0; i<nb_class; i++)
    confusion_matrix[i] = (int *) calloc(nb_class,sizeof(int));

  for (i=0; i<nb_class; i++) {
    sizet_total += test->class[i]->size;
    for (j=0; j<test->class[i]->size; j++) {
      vote = (double *) calloc(nb_class,sizeof(double));

      for (k=0; k<nb_class-1; k++) {
	for (l=k+1; l<nb_class; l++) {
	  ind = k*nb_class+l;

          // if using Levenshtein, need to normalize using std and mean
          if (sim == 0) {
	    mean = norm[ind][0];
	    std = norm[ind][1];
          // if using similarity learned with GESL, no need for normalization
          } else if (sim == 1) {
            mean = 0;
            std = 1;
          }

	  sizelk = train->class[k]->size;
	  sizell = train->class[l]->size;
	  
	  value = 0;
	  for (m=0; m<sizelk+sizell; m++) {
	    if (alpha[ind][m] != 0) {
	      if (m < sizelk) {
		if (sim == 0)
		  tmp = -levenshtein(test->class[i]->t[j],train->class[k]->t[m])*1.0;
		else if (sim == 1)
		  tmp = K_G(test->class[i]->t[j],train->class[k]->t[m],alph_size,ref_char,C[ind]);
		value += thresh((tmp-mean)/std)*alpha[ind][m];
	      } else {
		if (sim == 0)
		  tmp = -levenshtein(test->class[i]->t[j],train->class[l]->t[m-sizelk])*1.0;
		else if (sim == 1)
		  tmp = K_G(test->class[i]->t[j],train->class[l]->t[m-sizelk],alph_size,ref_char,C[ind]);
		value += thresh((tmp-mean)/std)*alpha[ind][m];
	      }
	    }
	  }
	  if (value >= 0)
	    vote[k] += 1;
	  else
	    vote[l] += 1;
	}
      }
      
      pred = 0;
      max = vote[0];
      for (m=1; m<nb_class; m++) {
	if (vote[m] > max) {
	  pred = m;
	  max = vote[m];
	}
      }

      free(vote);

      confusion_matrix[i][pred]++;
      
      if (pred != i) {
	nb_errors++;
	//fprintf(stderr,"Error: example %d, class %d, predicted %d\n",i*sizet_total+j,i,pred);
      }
    }
    
  }
  
  fprintf(stderr," Done.\nConfusion matrix:\n");
  for (i=0; i<nb_class; i++) {
    for (j=0; j<nb_class; j++) {
      fprintf(stderr,"%4d ",confusion_matrix[i][j]);
    }
    fprintf(stderr,"\n");
  }
  for (i=0; i<nb_class; i++)
    free(confusion_matrix[i]);
  free(confusion_matrix);
  fprintf(stderr,"Accuracy: %d/%d (%.2f\%)\n",sizet_total-nb_errors,sizet_total,(float)(sizet_total-nb_errors)/(sizet_total)*100);
}

