
/*
Copyright 2011 Aurélien Bellet
aurelien.bellet@univ-st-etienne.fr

This file is part of GESL.

GESL is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GESL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GESL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GESL_classify.h"

// learn a classifier on training data (using Levenshtein distance or matrices learned with GESL) and compute its performance on test data
int main (int argc, char *argv[]) {

  int i,j,p,q;
  int nb_class,alph_size;
  int sim; // edit similarity used: 0 (Levenshtein) or 1 (learned with GESL)
  double ***sim_mat_appr; // the similarity matrix(ces) on training data (one matrix per binary problem)
  double ***C; // the learned edit cost matrix (not needed if using Levenshtein)
  double **norm; // normalization constants (needed to normalize Levenshtein to [-1,1])
  double **alpha; // the classifier(s) learned with Balcan's learning rule
  double lambda; // regularization parameter on the L1-norm of alpha
  char ref_char; // reference character = character of smallest ASCII code appearing in the strings
  int err = 0;

  /* argument checking begins */
  if (argc < 8) {
    fprintf(stderr,"Wrong argument number.\nUsage: [training data dir] [test data dir] [nb class] [alphabet size] [ref char] [similarity type] [lambda] [matrix dir (optional)]\n");
    exit(-1);
  }
  nb_class = atoi(argv[3]);
  if (nb_class < 2) {
    err=1;
    fprintf(stderr,"Argument error: the number of class must be at least 2.\n");
  }
  alph_size = atoi(argv[4]);
  if (alph_size < 1) {
    err=1;
    fprintf(stderr,"Argument error: the alphabet size must be at least 1.\n");
  }
  ref_char = argv[5][0];
  sim = atoi(argv[6]);
  if (sim != 0 && sim != 1) {
    err=1;
    fprintf(stderr,"Argument error: the similarity type must be either 0 (Levenshtein) or 1 (GESL).\n");
  }
  lambda = atof(argv[7]);
  if (lambda < 0) {
    err=1;
    fprintf(stderr,"Argument error: lambda must be positive.\n");
  }
  if (err == 1) {
    fprintf(stderr,"Usage: [training data dir] [test data dir] [nb class] [alphabet size] [ref char] [similarity type] [lambda] [matrix dir (optional)]\n");
    exit(-1);
  }
  /* argument checking ends */

  sample *train; // training sample
  train = load_data(argv[1],nb_class); // load training data from file(s)
  sample *test; // test sample
  test = load_data(argv[2],nb_class); // load test data from file(s)

  // if using Levenshtein
  if (sim == 0) {
    sim_mat_appr = build_lev_mat(train); // compute the Levenshtein similarity matrix
    norm = normalize_mat(sim_mat_appr,train); // compute the normalization constants
  // else if using GESL-learned similarity
  } else if (sim == 1) {
    // load learned edit cost matrix(ces) from file(s)
    if (argc == 9)
      C = load_matrices(argv[8],alph_size,nb_class);
    else
      C = load_matrices(argv[1],alph_size,nb_class);
    sim_mat_appr = build_GESL_mat(train,alph_size,ref_char,C); // and compute the GESL similarity matrix
  }

  // build the linear program(s) used to infer the classifier(s)
  build_lp_mosek(train,sim_mat_appr,norm,argv[1],sim,lambda);

  // solve the LPs and get the resulting classifier(s) alpha
  alpha = solve_lp_mosek(argv[1],nb_class,train,sim,lambda);

  // print size of the learned classifier(s) alpha
  classifier_size(train,alpha);

  // classify test data using alpha
  classify_test(train,test,nb_class,norm,alpha,C,alph_size,ref_char,sim);

  return 0;

}

