
/*
Copyright 2011 Aurélien Bellet
aurelien.bellet@univ-st-etienne.fr

This file is part of GESL.

GESL is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GESL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GESL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GESL.h"

// learn edit cost matrix(ces) and save to file(s)
int main (int argc, char *argv[]) {

  int i,j;
  int nb_class,alph_size;
  int norm_type; // norm used in the optimization problem (1 for L1-norm, 2 for L2-norm)
  int pairing_strategy; // how the training examples are paired (0 for random, 1 for Levenshtein)
  char ref_char; // reference character = character of smallest ASCII code appearing in the strings
  double beta; // regularization parameter in the optimization problem
  double neta; // parameter for the "desired margin" in the optimization problem
  double alpha; // proportion of paired examples, must be in [0,1] (if alpha=1 all pairs are considered)
  int ****pairing;
  int err = 0;

  /* argument checking begins */
  if (argc < 10) {
    fprintf(stderr,"Wrong argument number.\nUsage: [data dir] [nb class] [alphabet size] [ref char] [norm type] [beta] [neta] [alpha] [pairing strategy] [output dir (optional)]\n");
    exit(-1);
  }
  nb_class = atoi(argv[2]);
  if (nb_class < 2) {
    err=1;
    fprintf(stderr,"Argument error: the number of class must be at least 2.\n");
  }
  alph_size = atoi(argv[3]);
  if (alph_size < 1) {
    err=1;
    fprintf(stderr,"Argument error: the alphabet size must be at least 1.\n");
  }
  ref_char = argv[4][0];
  norm_type = atoi(argv[5]);
  if (norm_type != 1 && norm_type != 2) {
    err=1;
    fprintf(stderr,"Argument error: the norm type must be either 1 (for L1-norm) or 2 (for L2-norm).\n");
  }
  beta = atof(argv[6]);
  neta = atof(argv[7]);
  if (beta < 0 || neta < 0) {
    err=1;
    fprintf(stderr,"Argument error: beta and neta must be positive.\n");
  }
  alpha = atof(argv[8]);
  if (alpha < 0 || alpha > 1) {
    err=1;
    fprintf(stderr,"Argument error: alpha must be in [0,1].\n");
  }
  pairing_strategy = atoi(argv[9]);
  if (pairing_strategy != 0 && pairing_strategy != 1) {
    err=1;
    fprintf(stderr,"Argument error: the pairing strategy must be either 0 (random pairing) or 1 (Levenshtein pairing).\n");
  }
  if (err == 1) {
    fprintf(stderr,"Usage: [data dir] [output dir] [nb class] [alphabet size] [ref char] [norm type] [beta] [neta] [alpha] [pairing strategy] [output dir (optional)]\n");
    exit(-1);
  }
  /* argument checking ends */

  sample *data; // training sample
  data = load_data(argv[1],nb_class); // load training data from file(s)

  // build the pairs
  if (pairing_strategy == 0)
    pairing = build_randpairing(data,alpha); // random pairing
  else
    pairing = build_levpairing(data,alpha,argv[1]); // levenshtein pairing
  
  // build the optimization problem(s)
  build_mosek(data,pairing,argv[1],beta,neta,alph_size,ref_char,alpha,norm_type);
  
  // solve the optimization problem(s) and save the matrix(ce) to file(s)
  if (argc == 11)
    solve_mosek(data,argv[1],argv[10],alph_size,beta,neta,alpha);
  else
    solve_mosek(data,argv[1],argv[1],alph_size,beta,neta,alpha);
  
  return 0;

}
